package ru.dbotthepony.kstarbound

import org.apache.logging.log4j.LogManager
import org.lwjgl.Version
import org.lwjgl.glfw.Callbacks.glfwFreeCallbacks
import org.lwjgl.glfw.GLFW.*
import org.lwjgl.glfw.GLFWErrorCallback
import org.lwjgl.opengl.GL46.*
import org.lwjgl.system.MemoryStack.stackPush
import org.lwjgl.system.MemoryUtil.NULL
import ru.dbotthepony.kstarbound.gl.*
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.render.Camera
import ru.dbotthepony.kstarbound.render.ChunkRenderer
import ru.dbotthepony.kstarbound.render.TileRenderer
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE_FF
import ru.dbotthepony.kstarbound.world.ChunkTile
import java.io.File
import kotlin.math.PI

private val LOGGER = LogManager.getLogger()
private const val TEST = false

var viewportWidth = 800
	private set
var viewportHeight = 600
	private set

var viewportMatrixGUI = updateViewportMatrixA()
	private set

var viewportMatrixGame = updateViewportMatrixB()
	private set

private fun updateViewportMatrixA(): Matrix4f {
	return Matrix4f.ortho(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 0.1f, 100f)
}

private fun updateViewportMatrixB(): Matrix4f {
	return Matrix4f.orthoDirect(0f, viewportWidth.toFloat(), 0f, viewportHeight.toFloat(), 1f, 100f)
}

var window = 0L
	private set

fun main() {
	LOGGER.info("Running LWJGL ${Version.getVersion()}")

	if (!TEST) {
		try {
			init()
			loop()
		} finally {
			if (window != NULL) {
				glfwFreeCallbacks(window)
				glfwDestroyWindow(window)
			}

			glfwTerminate()
			glfwSetErrorCallback(null)?.free()
		}
	} else {
		Starbound.addFilePath(File("./unpacked_assets/"))
		Starbound.loadTileDefinition("alienrock")
	}
}

private fun init() {
	GLFWErrorCallback.create { error, description ->
		LOGGER.error("LWJGL error {}: {}", error, description)
	}.set()

	check(glfwInit()) { "Unable to initialize GLFW" }

	glfwDefaultWindowHints()

	glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE)
	glfwWindowHint(GLFW_RESIZABLE, GLFW_TRUE)
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4)
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 6)
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE)

	window = glfwCreateWindow(viewportWidth, viewportHeight, "LWJGL Window!", NULL, NULL)
	require(window != NULL) { "Unable to create GLFW window" }

	glfwSetKeyCallback(window) { window, key, scancode, action, mods ->
		if (key == GLFW_KEY_ESCAPE || key == GLFW_RELEASE) {
			glfwSetWindowShouldClose(window, true)
		}
	}

	glfwSetFramebufferSizeCallback(window) { _, w, h ->
		viewportWidth = w
		viewportHeight = h
		viewportMatrixGUI = updateViewportMatrixA()
		viewportMatrixGame = updateViewportMatrixB()
		glViewport(0, 0, w, h)
	}

	val stack = stackPush()

	try {
		val pWidth = stack.mallocInt(1)
		val pHeight = stack.mallocInt(1)

		glfwGetWindowSize(window, pWidth, pHeight)

		val vidmode = glfwGetVideoMode(glfwGetPrimaryMonitor())!!

		glfwSetWindowPos(
			window,
			(vidmode.width() - pWidth[0]) / 2,
			(vidmode.height() - pHeight[0]) / 2
		)
	} finally {
		stack.close()
	}

	glfwMakeContextCurrent(window)

	// vsync
	glfwSwapInterval(1)

	glfwShowWindow(window)

	Starbound.addFilePath(File("./unpacked_assets/"))
	Starbound.loadTileMaterials()
}

private fun loop() {
	val state = GLStateTracker()
	val camera = Camera()

	// Set the clear color
	glClearColor(0.75f, 0.75f, 0.75f, 0.75f)

	state.blend = true
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

	val rock = ChunkTile(Starbound.getTileDefinition("alienrock")!!)
	val obsidian = ChunkTile(Starbound.getTileDefinition("obsidian")!!)
	val corruptdirt = ChunkTile(Starbound.getTileDefinition("corruptdirt")!!)
	val darkwood = ChunkTile(Starbound.getTileDefinition("darkwood")!!)
	val chunk = Starbound.world.setTile(Vector2i(2, 2), rock)

	chunk[3, 2] = rock
	chunk[4, 2] = rock
	chunk[4, 3] = rock
	chunk[4, 4] = rock
	chunk[3, 4] = rock
	chunk[5, 4] = rock
	chunk[2, 4] = obsidian
	chunk[2, 3] = obsidian
	chunk[1, 4] = corruptdirt
	chunk[1, 3] = corruptdirt
	chunk[1, 5] = darkwood
	chunk[2, 5] = darkwood
	chunk[3, 5] = darkwood
	chunk[1, 1] = darkwood
	chunk[2, 1] = darkwood
	chunk[3, 1] = darkwood

	for (x in 0 until 1) {
		for (y in 0 until 4) {
			//chunk[x, y] = ChunkTile(Starbound.loadTileDefinition("alienrock"))
		}
	}

	var x = 0
	var y = 0

	for (tile in Starbound.tilesAccess.values) {
		chunk[x++, y] = ChunkTile(tile)
		chunk[x++, y] = ChunkTile(tile)
		chunk[x++, y] = ChunkTile(tile)
		chunk[x++, y] = ChunkTile(tile)
		chunk[x++, y] = ChunkTile(tile)

		if (x >= 24) {
			x = 0
			y++
		}
	}

	val chunkRenderer = ChunkRenderer(state, chunk)
	chunkRenderer.tesselateStatic()
	chunkRenderer.uploadStatic()

	// Run the rendering loop until the user has attempted to close
	// the window or has pressed the ESCAPE key.
	while (!glfwWindowShouldClose(window)) {
		glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT) // clear the framebuffer

		state.matrixStack.clear(viewportMatrixGame.toMutableMatrix())

		// program.use()

		// val time = glfwGetTime()
		// program["globalColor"] = Uniform3f(sin(time).toFloat(), cos(time).toFloat(), sin(time).toFloat())
		// program["ourTexture"] = 0

		// texture.bind()
		// vao.bind()
		// ebo.bind()
		// glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, 0)
		// checkForGLError()

		//tileRenderer.renderPiece()

		//state.shaderVertexTexture.use()
		//state.shaderVertexTexture["_texture"] = 0
		////state.shaderVertexTexture["_transform"] = Vector3f.FORWARD.rotateAroundThis(PI * glfwGetTime())
		////state.shaderVertexTexture["_transform"] = state.matrixStack.push().scale((glfwGetTime() % 1000.0).toFloat(), (glfwGetTime() % 1000.0).toFloat(), (glfwGetTime() % 1000.0).toFloat()).last
		////state.shaderVertexTexture["_transform"] = Matrix4f.perspective(((glfwGetTime() * 16.0) % 180.0).toFloat(), 0.1f, 100f)
		//state.shaderVertexTexture["_transform"] = state.matrixStack.push().scale(x = 10f, y = 10f).translateWithScale(10f, 10f).last
		//texture.bind()
		//texture.textureMinFilter = GL_NEAREST
		//texture.textureMagFilter = GL_NEAREST
		//chunkRenderer.bind()
		//glDrawElements(GL_TRIANGLES, chunkRenderer.indexCount, GL_UNSIGNED_INT, 0)
		//checkForGLError()

		state.matrixStack.push().scale(x = 20f, y = 20f).translateWithScale(0f, 0f)
		chunkRenderer.render()
		//state.matrixStack.translateWithScale(18f)
		//chunkRenderer.render()

		glfwSwapBuffers(window) // swap the color buffers

		// Poll for window events. The key callback above will only be
		// invoked during this call.
		glfwPollEvents()
	}
}
