package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.math.Vector2i

data class ChunkTile(val def: TileDefinition) {
	companion object {
	}
}

interface IChunk {
	val pos: ChunkPos

	/**
	 * Возвращает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun get(x: Int, y: Int): ChunkTile?
	operator fun get(pos: Vector2i) = get(pos.x, pos.y)

	/**
	 * Относительная проверка находится ли координата вне границ чагка
	 */
	fun isOutside(x: Int, y: Int): Boolean {
		return x !in 0 until CHUNK_SIZE || y !in 0 until CHUNK_SIZE
	}

	/**
	 * Возвращает псевдослучайное Long для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomLongFor(x: Int, y: Int): Long {
		var long = x * 738548L + y * 2191293543L
		long = long xor 8339437585692L
		long = (long ushr 4) or (long shl 52)
		long *= 7848344324L
		long = (long ushr 12) or (long shl 44)
		return long
	}


	/**
	 * Возвращает псевдослучайное нормализированное Double для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomDoubleFor(x: Int, y: Int): Double {
		return (randomLongFor(x, y) / 9.223372036854776E18) / 2.0 + 0.5
	}


	/**
	 * Возвращает псевдослучайное Long для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomLongFor(pos: Vector2i) = randomLongFor(pos.x, pos.y)

	/**
	 * Возвращает псевдослучайное нормализированное Double для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomDoubleFor(pos: Vector2i) = randomDoubleFor(pos.x, pos.y)

	/**
	 * Возвращает итератор пар <Vector2i, Тайл?>
	 *
	 * Вектор имеет ОТНОСИТЕЛЬНЫЕ значения внутри самого чанка
	 */
	val posToTile: Iterator<Pair<Vector2i, ChunkTile?>> get() {
		return object : Iterator<Pair<Vector2i, ChunkTile?>> {
			private var x = 0
			private var y = 0

			private fun idx() = x + CHUNK_SIZE * y

			override fun hasNext(): Boolean {
				return idx() < CHUNK_SIZE * CHUNK_SIZE
			}

			override fun next(): Pair<Vector2i, ChunkTile?> {
				if (!hasNext()) {
					throw IllegalStateException("Already iterated everything!")
				}

				val tile = this@IChunk[x, y]
				val pos = Vector2i(x, y)

				x++

				if (x >= CHUNK_SIZE) {
					y++
					x = 0
				}

				return pos to tile
			}
		}
	}
}

interface IChunkSetter {
	/**
	 * Устанавливает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun set(x: Int, y: Int, tile: ChunkTile?)
}

interface IMutableChunk : IChunk, IChunkSetter

const val CHUNK_SHIFT = 6
const val CHUNK_SIZE = 1 shl CHUNK_SHIFT // 64
const val CHUNK_SIZE_FF = CHUNK_SIZE - 1

data class ChunkPos(val x: Int, val y: Int) {
	constructor(pos: Vector2i) : this(pos.x shr CHUNK_SHIFT, pos.y shr CHUNK_SHIFT)

	val firstBlock get() = Vector2i(x shl CHUNK_SHIFT, y shl CHUNK_SHIFT)
	val lastBlock get() = Vector2i(((x + 1) shl CHUNK_SHIFT) - 1, ((y + 1) shl CHUNK_SHIFT) - 1)
}

class Chunk(val world: World, override val pos: ChunkPos) : IMutableChunk {
	/**
	 * Хранит тайлы как x + y * CHUNK_SIZE
	 */
	val tiles = arrayOfNulls<ChunkTile>(CHUNK_SIZE * CHUNK_SIZE)

	override operator fun get(x: Int, y: Int): ChunkTile? {
		if (isOutside(x, y))
			return null

		return tiles[x or (y shl CHUNK_SHIFT)]
	}

	override operator fun set(x: Int, y: Int, tile: ChunkTile?) {
		if (isOutside(x, y))
			throw IndexOutOfBoundsException("Trying to set tile ${tile?.def?.materialName} at $x $y, but that is outside of chunk's range")

		tiles[x or (y shl CHUNK_SHIFT)] = tile
	}

	override fun randomLongFor(x: Int, y: Int): Long {
		return super.randomLongFor(x, y) xor world.seed
	}

	companion object {
		val EMPTY = object : IMutableChunk {
			override val pos = ChunkPos(0, 0)

			override fun get(x: Int, y: Int): ChunkTile? {
				return null
			}

			override fun set(x: Int, y: Int, tile: ChunkTile?) {

			}
		}
	}
}
